Unit PerceptronCore;
interface
const
  InOutVectorSize = 64;
  ALayerSize = 128;
type
(* Artificial Cells *)
  cell = object
	public
	  firing : boolean;		(* output *)
	  procedure init(initTag : byte);
	  function  getTag : byte;
	private
	  tag    : byte;
  end;
  pcell = ^cell;
  connexion = record
	from : pcell;
	weight : single
  end;
  ConnexionVector = array[1..ALayerSize] of connexion;
  RegularCell = object(cell)
	public
	  procedure init(initTag : byte);
	  procedure AddConnexion(fr : pcell; w : single);
	  procedure CalcOutput;
	  procedure SpyStatus(var e, t : single;
						  var sc : ConnexionVector;
						  var cn : integer);
	  procedure done;
	private
	  Em, threshold : single;
	  c : ConnexionVector;
	  cnum : integer;
  end;
  pRegularCell = ^RegularCell;
  BasicPerceptron = object(RegularCell)
	public
	  procedure AdjustWeights(teacher : boolean; delta : single);
  end;
  pBasicPerceptron = ^BasicPerceptron;

(* Input/Output Pattern *)
  InOutVectorIndex = 1..InOutVectorSize;
  InOutVector = set of InOutVectorIndex;
  pattern = record
	vec, typ : InOutVector
  end;
(* Peeping Neuron *)
  ResultVector = array[InOutVectorIndex] of integer;

procedure SetUpSLayer(v : InOutVector);
procedure SetUpALayer;
procedure SetUpRLayer;
procedure SetInputVector(v : InOutVector);
procedure Calc;
procedure TeachAPattern(var p : pattern);
procedure TeachAPatternWithResult(var p : pattern; var r : ResultVector);
function  PeepCell(isRLayer : boolean; n : integer) : pRegularCell;
procedure GetOutputVector(var v : InOutVector);
procedure DestroyLayers;

var
  SLayerCellNum,
  ALayerCellNum,
  RLayerCellNum : integer;
  epsilon : single;

implementation

const initEpsilon  = 0.05;
var
  SLayer : array[InOutVectorIndex] of pcell;
  ALayer : array[1..ALayerSize] of pRegularCell;
  RLayer : array[InOutVectorIndex] of pBasicPerceptron;

function Rand : single;
begin
  Rand := 1 - random * 2
end;

procedure cell.init(initTag : byte);
begin
  firing := false;
  tag := initTag
end;

function cell.getTag : byte;
begin
  getTag := tag
end;

procedure RegularCell.init(initTag : byte);
begin
  Em := 0;
  threshold := Rand;
  cnum := 0;		(* no connexion *)
  inherited init(initTag)
end;

procedure RegularCell.AddConnexion(fr : pcell; w : single);
begin
  inc(cnum);
  with c[cnum] do begin
	from := fr;
	weight := w
  end
end;

procedure RegularCell.CalcOutput;
var i : integer;
begin
  Em := 0;
  for i := 1 to cnum do
	with c[i] do
	  if from^.firing then Em := Em + weight;
  firing := Em > threshold
end;

procedure RegularCell.SpyStatus(var e, t : single;
						  var sc : ConnexionVector;
						  var cn : integer);
begin
  e := Em;
  t := threshold;
  sc:= c;
  cn:= cnum
end;

procedure RegularCell.done;
begin
  cnum := 0
end;

procedure BasicPerceptron.AdjustWeights(teacher : boolean; delta : single);
var i : integer;
begin
(* This method should be called if firing <> teacher *)
  if firing > teacher then delta := -delta;
	 (* The weight must be decreased in this case.
		Corresponding to making a product of current output
		and the teacher input. *)
  for i := 1 to cnum do
	with c[i] do
	  if from^.firing then weight := weight + delta;
  threshold := threshold - delta
end;

procedure SetUpSLayer(v : InOutVector);
var i : InOutVectorIndex;
begin
  for i := 1 to SLayerCellNum do begin
	new(SLayer[i]);
	SLayer[i]^.init(i);
	SLayer[i]^.firing := (i in v)
  end
end;

procedure SetUpALayer;
var
  i : 1..ALayerSize;
  j : InOutVectorIndex;
begin
  for i := 1 to ALayerCellNum do begin
	new(ALayer[i]);
	with ALayer[i]^ do begin
	  init(i);
	  for j := 1 to SLayerCellNum do AddConnexion(SLayer[j], Rand)
	end
  end
end;

procedure SetUpRLayer;
var i, j : integer;
begin
  for i := 1 to RLayerCellNum do begin
	new(RLayer[i]);
	with RLayer[i]^ do begin
	  init(i);
	  if ALayerCellNum > 0 then
		for j := 1 to ALayerCellNum do
		  AddConnexion(ALayer[j], Rand)
	  else
		for j := 1 to SLayerCellNum do
		  AddConnexion(SLayer[j], Rand)
	end
  end
end;

procedure DestroyLayers;
var i : integer;
begin
  for i := 1 to RLayerCellNum do begin
	RLayer[i]^.done;
	dispose(RLayer[i])
  end;
  filldword(RLayer, sizeof(RLayer) div 4, 0);
  RLayerCellNum := 0;
  for i := 1 to ALayerCellNum do begin
	ALayer[i]^.done;
	dispose(ALayer[i])
  end;
  filldword(ALayer, sizeof(ALayer) div 4, 0);
  ALayerCellNum := 0;
  for i := 1 to SLayerCellNum do dispose(SLayer[i]);
  filldword(SLayer, sizeof(SLayer) div 4, 0);
  SLayerCellNum := 0
end;

procedure SetInputVector(v : InOutVector);
var i : integer;
begin
  for i := 1 to SLayerCellNum do SLayer[i]^.firing := (i in v)
end;
	
procedure Calc;
var i : integer;
begin
  for i := 1 to ALayerCellNum do ALayer[i]^.CalcOutput;
  for i := 1 to RLayerCellNum do RLayer[i]^.CalcOutput
end;

procedure TeachAPattern(var p : pattern);
var j : integer;
	CorrectAnswer : boolean;
begin
  SetInputVector(p.vec);
  Calc;
  for j := 1 to RLayerCellNum do begin
	CorrectAnswer := (j in p.typ);
	with RLayer[j]^ do
	  if firing <> CorrectAnswer then
		AdjustWeights(CorrectAnswer, epsilon)
  end
end;

procedure TeachAPatternWithResult(var p : pattern; var r : ResultVector);
var j : integer;
	CorrectAnswer : boolean;
begin
  SetInputVector(p.vec);
  Calc;
  for j := 1 to RLayerCellNum do begin
	CorrectAnswer := (j in p.typ);
	with RLayer[j]^ do begin
	  if firing <> CorrectAnswer then
		AdjustWeights(CorrectAnswer, epsilon);
	  r[j] := ord(CorrectAnswer) - ord(firing)
	end
  end
end;

function  PeepCell(isRLayer : boolean; n : integer) : pRegularCell;
begin
  PeepCell := nil;
  if isRLayer then begin
	if n <= RLayerCellNum then PeepCell := RLayer[n]
  end else
	if n <= ALayerCellNum then PeepCell := ALayer[n]
end;

procedure GetOutputVector(var v : InOutVector);
var i : integer;
begin
  v := [];
  for i := 1 to RLayerCellNum do
	if RLayer[i]^.firing then include(v, i)
end;

begin
  SLayerCellNum := 0;
  ALayerCellNum := 0;
  RLayerCellNum := 0;
  filldword(SLayer, sizeof(SLayer) div 4, 0);
  filldword(ALayer, sizeof(ALayer) div 4, 0);
  filldword(RLayer, sizeof(RLayer) div 4, 0);
  epsilon := initEpsilon
end.
